//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/DetectorEditor.cpp
//! @brief     Implements class DetectorEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/DetectorEditor.h"
#include "GUI/Model/Detector/DetectorItem.h"
#include "GUI/Model/Detector/ResolutionFunctionItems.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/View/Numeric/ComboUtil.h"
#include "GUI/View/Numeric/DoubleSpinBox.h"
#include "GUI/View/Numeric/NumWidgetUtil.h"
#include "GUI/View/Numeric/SafeSpinBox.h"
#include <QFormLayout>

DetectorEditor::DetectorEditor(QWidget* parent, GISASInstrumentItem* instrItem)
    : CollapsibleGroupBox("Detector parameters", parent, instrItem->expandDetector)
{
    auto* layout = new QVBoxLayout;
    body()->setLayout(layout);

    DetectorItem* detectorItem = instrItem->detectorItem();
    ASSERT(detectorItem);

    auto* xyrow = new QHBoxLayout;
    layout->addLayout(xyrow);

    //... x-axis controls

    auto* xAxisGroupBox = new StaticGroupBox("Phi axis", this);
    xyrow->addWidget(xAxisGroupBox);

    auto* xAxisFormLayout = new QFormLayout;
    xAxisGroupBox->body()->setLayout(xAxisFormLayout);

    auto* xAxisNbinsSpinBox = new SafeSpinBox;
    xAxisFormLayout->addRow("# bins:", xAxisNbinsSpinBox);
    xAxisNbinsSpinBox->setRange(1, 65536);
    xAxisNbinsSpinBox->setValue(detectorItem->nPhi());
    connect(xAxisNbinsSpinBox, &QSpinBox::valueChanged, [this, detectorItem](int newValue) {
        detectorItem->setNPhi(newValue);
        emit dataChanged();
    });

    auto* phiMinSpinBox = new DoubleSpinBox(detectorItem->phiMin());
    xAxisFormLayout->addRow("Phi min (deg):", phiMinSpinBox);
    connect(phiMinSpinBox, &DoubleSpinBox::baseValueChanged, [this, detectorItem](double newValue) {
        detectorItem->phiMin().setValue(newValue);
        emit dataChanged();
    });

    auto* phiMaxSpinBox = new DoubleSpinBox(detectorItem->phiMax());
    xAxisFormLayout->addRow("Phi max (deg):", phiMaxSpinBox);
    connect(phiMaxSpinBox, &DoubleSpinBox::baseValueChanged, [this, detectorItem](double newValue) {
        detectorItem->phiMax().setValue(newValue);
        emit dataChanged();
    });

    //... y-axis controls

    auto* yAxisGroupBox = new StaticGroupBox("Alpha axis", this);
    xyrow->addWidget(yAxisGroupBox);

    auto* yAxisFormLayout = new QFormLayout;
    yAxisGroupBox->body()->setLayout(yAxisFormLayout);

    auto* yAxisNbinsSpinBox = new SafeSpinBox;
    yAxisFormLayout->addRow("# bins:", yAxisNbinsSpinBox);
    yAxisNbinsSpinBox->setRange(1, 65536);
    yAxisNbinsSpinBox->setValue(detectorItem->nAlpha());
    connect(yAxisNbinsSpinBox, &QSpinBox::valueChanged, [this, detectorItem](int newValue) {
        detectorItem->setNAlpha(newValue);
        emit dataChanged();
    });

    auto* alphaMinSpinBox = new DoubleSpinBox(detectorItem->alphaMin());
    xAxisFormLayout->addRow("Alpha min (deg):", alphaMinSpinBox);
    connect(alphaMinSpinBox, &DoubleSpinBox::baseValueChanged, [this, detectorItem](double newValue) {
        detectorItem->alphaMin().setValue(newValue);
        emit dataChanged();
    });

    auto* alphaMaxSpinBox = new DoubleSpinBox(detectorItem->alphaMax());
    xAxisFormLayout->addRow("Alpha max (deg):", alphaMaxSpinBox);
    connect(alphaMaxSpinBox, &DoubleSpinBox::baseValueChanged, [this, detectorItem](double newValue) {
        detectorItem->alphaMax().setValue(newValue);
        emit dataChanged();
    });

    //... resolution controls

    auto* resolutionForm = new StaticGroupBox("Resolution function", this);
    xyrow->addWidget(resolutionForm);

    auto resolutionLayout = new QFormLayout;
    resolutionForm->body()->setLayout(resolutionLayout);
    resolutionLayout->setFieldGrowthPolicy(QFormLayout::FieldsStayAtSizeHint);

    auto updateResolutionForm = [parent = this, layout = resolutionLayout,
                                 item = detectorItem]() -> void {
        while (layout->rowCount() > 1)
            layout->removeRow(1);

        auto* resFunction = item->resolutionFunctionSelection().currentItem();
        if (auto* p = dynamic_cast<ResolutionFunction2DGaussianItem*>(resFunction)) {
            auto* sigmaXSpinBox = GUI::Util::createDoubleSpinBoxRow(layout, p->sigmaX());
            auto* sigmaYSpinBox = GUI::Util::createDoubleSpinBoxRow(layout, p->sigmaY());

            connect(sigmaXSpinBox, &DoubleSpinBox::baseValueChanged, [parent, p](double newValue) {
                p->setSigmaX(newValue);
                emit parent->dataChanged();
            });

            connect(sigmaYSpinBox, &DoubleSpinBox::baseValueChanged, [parent, p](double newValue) {
                p->setSigmaY(newValue);
                emit parent->dataChanged();
            });
        }
    };

    auto* typeCombo = GUI::Util::createComboBoxFromProperty(
        detectorItem->resolutionFunctionSelection(),
        [this, updateResolutionForm](int) {
            updateResolutionForm();
            emit dataChanged();
        },
        false);
    resolutionLayout->addRow("Type:", typeCombo);

    updateResolutionForm();
}
