//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/SampleDesigner/SampleView.cpp
//! @brief     Implements class SampleView
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/SampleDesigner/SampleView.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Model/Sample/ItemWithParticles.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/Sample/SampleItem.h"
#include "GUI/View/SampleDesigner/SampleEditor.h"
#include "GUI/View/SampleDesigner/SampleListing.h"
#include "GUI/View/SampleDesigner/ScriptPanel.h"
#include "GUI/View/SampleView/RealspacePanel.h"
#include "GUI/View/SampleView/RealspaceWidget.h"
#include "GUI/View/Widget/ApplicationSettings.h"
#include "GUI/View/Widget/StyledToolbar.h"

#include <QBoxLayout>
#include <QCheckBox>
#include <QDockWidget>
#include <QMenu>
#include <QMessageBox>
#include <QToolButton>
#include <QWidgetAction>

SampleView::SampleView(QWidget* parent, ProjectDocument* document)
    : QWidget(parent)
    , m_document(document)
{
    auto* layout = new QVBoxLayout(this);

    // Top toolbar with action buttons "new", "import", "example"
    auto* toolbar = new StyledToolbar;
    layout->addWidget(toolbar);
    toolbar->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);

    // Everything below
    auto* hLayout = new QHBoxLayout;
    layout->addLayout(hLayout);

    //... Left margin: sample listing

    m_listing = new SampleListing(this, m_document);
    hLayout->addWidget(m_listing);
    m_listing->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);

    //... Right block: sample editor and bottom panels

    auto* vLayout = new QVBoxLayout;
    hLayout->addLayout(vLayout);

    auto* editor = new SampleEditor(this, document);
    vLayout->addWidget(editor);

    //... Below central widget: realspace and script panels

    auto* hLayout2 = new QHBoxLayout;
    vLayout->addLayout(hLayout2);

    m_realspacePanel = new RealspacePanel(this);
    hLayout2->addWidget(m_realspacePanel);

    auto* scriptPanel = new ScriptPanel(this);
    hLayout2->addWidget(scriptPanel);

    //... Finish

    toolbar->addAction(m_listing->newSampleAction());
#ifdef BORNAGAIN_PYTHON
    toolbar->addAction(m_listing->importSampleAction());
#endif
    toolbar->addAction(m_listing->chooseFromLibraryAction());
    if (auto* btn = dynamic_cast<QToolButton*>(
            toolbar->widgetForAction(m_listing->chooseFromLibraryAction())))
        btn->setPopupMode(QToolButton::InstantPopup);

    connect(m_document, &ProjectDocument::singleSampleModeChanged, this,
            &SampleView::updateSingleSampleMode);

    connect(m_listing, &SampleListing::currentSampleChanged, editor,
            &SampleEditor::setCurrentSample);

    connect(m_listing, &SampleListing::currentSampleChanged,
            [&](SampleItem* m) { onRequestViewInRealspace(m); });

    connect(m_listing, &SampleListing::currentSampleChanged, scriptPanel,
            &ScriptPanel::setCurrentSample);

    connect(editor, &SampleEditor::modified, scriptPanel, &ScriptPanel::onSampleModified);

    connect(editor, &SampleEditor::modified, m_listing, &SampleListing::onSampleModified);

    connect(editor, &SampleEditor::requestViewInRealspace, this,
            &SampleView::onRequestViewInRealspace);

    connect(editor, &SampleEditor::aboutToRemoveItem, this, &SampleView::onAboutToRemoveItem);

    connect(editor, &SampleEditor::requestCreateNewSample, m_listing->newSampleAction(),
            &QAction::trigger, Qt::QueuedConnection);

    connect(editor, &SampleEditor::modified, m_realspacePanel->widget(),
            &RealspaceWidget::updateScene);

    connect(editor, &SampleEditor::modified, m_document, &ProjectDocument::setModified,
            Qt::UniqueConnection);
}

void SampleView::updateSingleSampleMode()
{
    // m_docks->setDockVisible(SAMPLE_LIST, !m_document->singleSampleMode());
}

void SampleView::onRequestViewInRealspace(Item3D* item)
{
    if (!item) {
        m_realspacePanel->widget()->clearDisplay();
        return;
    }

    // m_docks->setDockVisible(REALSPACEPANEL);
    m_realspacePanel->widget()->setDisplayedItem(m_listing->currentSampleItem(), item);
}

void SampleView::onAboutToRemoveItem(Item3D* item)
{
    // If an item shall be deleted from the sample, check whether Realspace is affected.
    // Its current item could be item itself, or it could be a child of item.

    // -- for convenience access later on
    auto* widget = m_realspacePanel->widget();
    Item3D* displayed = widget->displayedItem();

    if (dynamic_cast<SampleItem*>(item)) {
        // It is a sample. In this case: always reset
        widget->resetScene();

    } else if (auto* i = dynamic_cast<LayerItem*>(item)) {
        if (item == widget->displayedItem() || i->layoutItems().contains(displayed)
            || i->itemsWithParticles().contains(displayed))
            widget->resetScene();

    } else if (auto* i = dynamic_cast<ParticleLayoutItem*>(item)) {
        if (item == widget->displayedItem() || i->containedItemsWithParticles().contains(displayed))
            widget->resetScene();

    } else if (auto* i = dynamic_cast<ItemWithParticles*>(item)) {
        if (item == displayed || i->containedItemsWithParticles().contains(displayed))
            widget->resetScene();
    }
}

void SampleView::fillViewMenu(QMenu* menu)
{
    // m_docks->addDockActionsToMenu(menu);

    menu->addSeparator();

    auto* actionSampleMode = new QWidgetAction(menu);
    actionSampleMode->setText("Use single sample mode");
    auto* sampleModeWidget = new QCheckBox("Use single sample mode", menu);
    sampleModeWidget->setChecked(m_document->singleSampleMode());
    connect(sampleModeWidget, &QCheckBox::toggled, this, &SampleView::onSingleSampleModeChanged);
    actionSampleMode->setCheckable(true);
    actionSampleMode->setDefaultWidget(sampleModeWidget);
    menu->addAction(actionSampleMode);
}

void SampleView::showEvent(QShowEvent*)
{
    if (!m_listing->currentSampleItem())
        m_listing->restoreSelection();
}

void SampleView::onSingleSampleModeChanged(bool newState)
{
    if (newState) {
        if (m_document->sampleModel()->sampleItems().size() > 1) {
            QMessageBox::warning(this, "Select single sample mode",
                                 "This project already contains more than one sample. Changing "
                                 "this setting is not possible.");
            return;
        }

        m_document->setSingleSampleMode(true);
    } else
        m_document->setSingleSampleMode(false);
    if (gProjectDocument.has_value()) {
        appSettings->setDefaultIsSingleSampleMode(gProjectDocument.value()->singleSampleMode());
    }
}
